from __future__ import annotations

import json

from typing import Dict, Optional, Union


from samtranslator.schema.common import BaseModel, LenientBaseModel
from samtranslator.schema import (
    aws_serverless_simpletable,
    aws_serverless_application,
    aws_serverless_connector,
    aws_serverless_function,
    aws_serverless_statemachine,
    aws_serverless_layerversion,
    aws_serverless_api,
    aws_serverless_httpapi,
    any_cfn_resource,
)


class Globals(BaseModel):
    Function: Optional[aws_serverless_function.Globals]
    Api: Optional[aws_serverless_api.Globals]
    HttpApi: Optional[aws_serverless_httpapi.Globals]
    SimpleTable: Optional[aws_serverless_simpletable.Globals]


class Model(LenientBaseModel):
    Globals: Optional[Globals]
    Resources: Dict[
        str,
        Union[
            aws_serverless_connector.Resource,
            aws_serverless_function.Resource,
            aws_serverless_simpletable.Resource,
            aws_serverless_statemachine.Resource,
            aws_serverless_layerversion.Resource,
            aws_serverless_api.Resource,
            aws_serverless_httpapi.Resource,
            aws_serverless_application.Resource,
            any_cfn_resource.Resource,
        ],
    ]


def main() -> None:
    obj = Model.schema()

    # http://json-schema.org/understanding-json-schema/reference/schema.html#schema
    # https://github.com/pydantic/pydantic/issues/1478
    # Validated in https://github.com/aws/serverless-application-model/blob/5c82f5d2ae95adabc9827398fba8ccfc3dbe101a/tests/schema/test_validate_schema.py#L91
    obj["$schema"] = "http://json-schema.org/draft-04/schema#"

    print(json.dumps(obj, indent=2))


if __name__ == "__main__":
    main()



from samtranslator.model import PropertyType, Resource
from samtranslator.model.types import IS_DICT, list_of, IS_STR
from samtranslator.model.intrinsics import fnGetAtt, ref


class CognitoUserPool(Resource):
    resource_type = "AWS::Cognito::UserPool"
    property_types = {
        "AccountRecoverySetting": PropertyType(False, IS_DICT),
        "AdminCreateUserConfig": PropertyType(False, IS_DICT),
        "AliasAttributes": PropertyType(False, list_of(IS_STR)),
        "AutoVerifiedAttributes": PropertyType(False, list_of(IS_STR)),
        "DeviceConfiguration": PropertyType(False, IS_DICT),
        "EmailConfiguration": PropertyType(False, IS_DICT),
        "EmailVerificationMessage": PropertyType(False, IS_STR),
        "EmailVerificationSubject": PropertyType(False, IS_STR),
        "EnabledMfas": PropertyType(False, list_of(IS_STR)),
        "LambdaConfig": PropertyType(False, IS_DICT),
        "MfaConfiguration": PropertyType(False, IS_STR),
        "Policies": PropertyType(False, IS_DICT),
        "Schema": PropertyType(False, list_of(dict)),
        "SmsAuthenticationMessage": PropertyType(False, IS_STR),
        "SmsConfiguration": PropertyType(False, IS_DICT),
        "SmsVerificationMessage": PropertyType(False, IS_STR),
        "UserAttributeUpdateSettings": PropertyType(False, IS_DICT),
        "UsernameAttributes": PropertyType(False, list_of(IS_STR)),
        "UsernameConfiguration": PropertyType(False, IS_DICT),
        "UserPoolAddOns": PropertyType(False, list_of(dict)),
        "UserPoolName": PropertyType(False, IS_STR),
        "UserPoolTags": PropertyType(False, IS_DICT),
        "VerificationMessageTemplate": PropertyType(False, IS_DICT),
    }

    runtime_attrs = {
        "name": lambda self: ref(self.logical_id),
        "arn": lambda self: fnGetAtt(self.logical_id, "Arn"),
        "provider_name": lambda self: fnGetAtt(self.logical_id, "ProviderName"),
        "provider_url": lambda self: fnGetAtt(self.logical_id, "ProviderURL"),
    }



from typing import Optional, Dict, Any, List, Union
from samtranslator.model import PropertyType, Resource, PassThroughProperty
from samtranslator.model.types import IS_DICT, is_type, one_of, IS_STR, list_of, any_type
from samtranslator.model.intrinsics import fnGetAtt, ref
from samtranslator.utils.types import Intrinsicable


class LambdaFunction(Resource):
    resource_type = "AWS::Lambda::Function"
    property_types = {
        "Code": PropertyType(True, IS_DICT),
        "PackageType": PropertyType(False, IS_STR),
        "DeadLetterConfig": PropertyType(False, IS_DICT),
        "Description": PropertyType(False, IS_STR),
        "FunctionName": PropertyType(False, IS_STR),
        "Handler": PropertyType(False, IS_STR),
        "MemorySize": PropertyType(False, is_type(int)),
        "Role": PropertyType(False, IS_STR),
        "Runtime": PropertyType(False, IS_STR),
        "Timeout": PropertyType(False, is_type(int)),
        "VpcConfig": PropertyType(False, IS_DICT),
        "Environment": PropertyType(False, IS_DICT),
        "Tags": PropertyType(False, list_of(IS_DICT)),
        "TracingConfig": PropertyType(False, IS_DICT),
        "KmsKeyArn": PropertyType(False, one_of(IS_DICT, IS_STR)),
        "Layers": PropertyType(False, list_of(one_of(IS_STR, IS_DICT))),
        "ReservedConcurrentExecutions": PropertyType(False, any_type()),
        "FileSystemConfigs": PropertyType(False, list_of(IS_DICT)),
        "CodeSigningConfigArn": PropertyType(False, IS_STR),
        "ImageConfig": PropertyType(False, IS_DICT),
        "Architectures": PropertyType(False, list_of(one_of(IS_STR, IS_DICT))),
        "SnapStart": PropertyType(False, IS_DICT),
        "EphemeralStorage": PropertyType(False, IS_DICT),
        "RuntimeManagementConfig": PropertyType(False, IS_DICT),
    }

    Code: Dict[str, Any]
    PackageType: Optional[str]
    DeadLetterConfig: Optional[Dict[str, Any]]
    Description: Optional[Intrinsicable[str]]
    FunctionName: Optional[Intrinsicable[str]]
    Handler: Optional[str]
    MemorySize: Optional[Intrinsicable[int]]
    Role: Optional[Intrinsicable[str]]
    Runtime: Optional[str]
    Timeout: Optional[Intrinsicable[int]]
    VpcConfig: Optional[Dict[str, Any]]
    Environment: Optional[Dict[str, Any]]
    Tags: Optional[List[Dict[str, Any]]]
    TracingConfig: Optional[Dict[str, Any]]
    KmsKeyArn: Optional[Intrinsicable[str]]
    Layers: Optional[List[Any]]
    ReservedConcurrentExecutions: Optional[Any]
    FileSystemConfigs: Optional[Dict[str, Any]]
    CodeSigningConfigArn: Optional[Intrinsicable[str]]
    ImageConfig: Optional[Dict[str, Any]]
    Architectures: Optional[List[Any]]
    SnapStart: Optional[Dict[str, Any]]
    EphemeralStorage: Optional[Dict[str, Any]]
    RuntimeManagementConfig: Optional[Dict[str, Any]]

    runtime_attrs = {"name": lambda self: ref(self.logical_id), "arn": lambda self: fnGetAtt(self.logical_id, "Arn")}


class LambdaVersion(Resource):
    resource_type = "AWS::Lambda::Version"
    property_types = {
        "CodeSha256": PropertyType(False, IS_STR),
        "Description": PropertyType(False, IS_STR),
        "FunctionName": PropertyType(True, one_of(IS_STR, IS_DICT)),
        "RuntimeManagementConfig": PropertyType(False, IS_DICT),
    }

    runtime_attrs = {
        "arn": lambda self: ref(self.logical_id),
        "version": lambda self: fnGetAtt(self.logical_id, "Version"),
    }


class LambdaAlias(Resource):
    resource_type = "AWS::Lambda::Alias"
    property_types = {
        "Description": PropertyType(False, IS_STR),
        "Name": PropertyType(False, IS_STR),
        "FunctionName": PropertyType(True, one_of(IS_STR, IS_DICT)),
        "FunctionVersion": PropertyType(True, one_of(IS_STR, IS_DICT)),
        "ProvisionedConcurrencyConfig": PropertyType(False, IS_DICT),
    }

    runtime_attrs = {"arn": lambda self: ref(self.logical_id)}


class LambdaEventSourceMapping(Resource):
    resource_type = "AWS::Lambda::EventSourceMapping"
    property_types = {
        "BatchSize": PropertyType(False, is_type(int)),
        "Enabled": PropertyType(False, is_type(bool)),
        "EventSourceArn": PropertyType(False, IS_STR),
        "FunctionName": PropertyType(True, IS_STR),
        "MaximumBatchingWindowInSeconds": PropertyType(False, is_type(int)),
        "MaximumRetryAttempts": PropertyType(False, is_type(int)),
        "BisectBatchOnFunctionError": PropertyType(False, is_type(bool)),
        "MaximumRecordAgeInSeconds": PropertyType(False, is_type(int)),
        "DestinationConfig": PropertyType(False, IS_DICT),
        "ParallelizationFactor": PropertyType(False, is_type(int)),
        "StartingPosition": PropertyType(False, IS_STR),
        "StartingPositionTimestamp": PassThroughProperty(False),
        "Topics": PropertyType(False, is_type(list)),
        "Queues": PropertyType(False, is_type(list)),
        "SourceAccessConfigurations": PropertyType(False, is_type(list)),
        "TumblingWindowInSeconds": PropertyType(False, is_type(int)),
        "FunctionResponseTypes": PropertyType(False, is_type(list)),
        "SelfManagedEventSource": PropertyType(False, IS_DICT),
        "FilterCriteria": PropertyType(False, IS_DICT),
        "AmazonManagedKafkaEventSourceConfig": PropertyType(False, IS_DICT),
        "SelfManagedKafkaEventSourceConfig": PropertyType(False, IS_DICT),
        "ScalingConfig": PropertyType(False, IS_DICT),
    }

    runtime_attrs = {"name": lambda self: ref(self.logical_id)}


class LambdaPermission(Resource):
    resource_type = "AWS::Lambda::Permission"
    property_types = {
        "Action": PropertyType(True, IS_STR),
        "FunctionName": PropertyType(True, IS_STR),
        "Principal": PropertyType(True, IS_STR),
        "SourceAccount": PropertyType(False, IS_STR),
        "SourceArn": PropertyType(False, IS_STR),
        "EventSourceToken": PropertyType(False, IS_STR),
        "FunctionUrlAuthType": PropertyType(False, IS_STR),
    }


class LambdaEventInvokeConfig(Resource):
    resource_type = "AWS::Lambda::EventInvokeConfig"
    property_types = {
        "DestinationConfig": PropertyType(False, IS_DICT),
        "FunctionName": PropertyType(True, IS_STR),
        "MaximumEventAgeInSeconds": PropertyType(False, is_type(int)),
        "MaximumRetryAttempts": PropertyType(False, is_type(int)),
        "Qualifier": PropertyType(True, IS_STR),
    }


class LambdaLayerVersion(Resource):
    """Lambda layer version resource"""

    resource_type = "AWS::Lambda::LayerVersion"
    property_types = {
        "Content": PropertyType(True, IS_DICT),
        "Description": PropertyType(False, IS_STR),
        "LayerName": PropertyType(False, IS_STR),
        "CompatibleArchitectures": PropertyType(False, list_of(one_of(IS_STR, IS_DICT))),
        "CompatibleRuntimes": PropertyType(False, list_of(one_of(IS_STR, IS_DICT))),
        "LicenseInfo": PropertyType(False, IS_STR),
    }

    Content: Dict[str, Any]
    Description: Optional[Intrinsicable[str]]
    LayerName: Optional[Intrinsicable[str]]
    CompatibleArchitectures: Optional[List[Union[str, Dict[str, Any]]]]
    CompatibleRuntimes: Optional[List[Union[str, Dict[str, Any]]]]
    LicenseInfo: Optional[Intrinsicable[str]]

    runtime_attrs = {"name": lambda self: ref(self.logical_id), "arn": lambda self: fnGetAtt(self.logical_id, "Arn")}


class LambdaUrl(Resource):
    resource_type = "AWS::Lambda::Url"
    property_types = {
        "TargetFunctionArn": PropertyType(True, one_of(IS_STR, IS_DICT)),
        "AuthType": PropertyType(True, IS_STR),
        "Cors": PropertyType(False, IS_DICT),
    }



from typing import Optional, List

from typing_extensions import Literal

from samtranslator.schema.common import PassThrough, BaseModel, get_prop

resourcereference = get_prop("sam-property-connector-resourcereference")
properties = get_prop("sam-resource-connector")


class ResourceReference(BaseModel):
    Id: Optional[str] = resourcereference("Id")
    Arn: Optional[PassThrough] = resourcereference("Arn")
    Name: Optional[PassThrough] = resourcereference("Name")
    Qualifier: Optional[PassThrough] = resourcereference("Qualifier")
    QueueUrl: Optional[PassThrough] = resourcereference("QueueUrl")
    ResourceId: Optional[PassThrough] = resourcereference("ResourceId")
    RoleName: Optional[PassThrough] = resourcereference("RoleName")
    Type: Optional[str] = resourcereference("Type")


class Properties(BaseModel):
    Source: ResourceReference = properties("Source")
    Destination: ResourceReference = properties("Destination")
    Permissions: List[Literal["Read", "Write"]] = properties("Permissions")


class Resource(BaseModel):
    Type: Literal["AWS::Serverless::Connector"]
    Properties: Properties



from samtranslator.model import PropertyType, Resource
from samtranslator.model.intrinsics import ref
from samtranslator.model.types import IS_DICT, is_type, one_of, IS_STR


class CodeDeployApplication(Resource):
    resource_type = "AWS::CodeDeploy::Application"
    property_types = {"ComputePlatform": PropertyType(False, one_of(IS_STR, IS_DICT))}

    runtime_attrs = {"name": lambda self: ref(self.logical_id)}


class CodeDeployDeploymentGroup(Resource):
    resource_type = "AWS::CodeDeploy::DeploymentGroup"
    property_types = {
        "AlarmConfiguration": PropertyType(False, IS_DICT),
        "ApplicationName": PropertyType(True, one_of(IS_STR, IS_DICT)),
        "AutoRollbackConfiguration": PropertyType(False, IS_DICT),
        "DeploymentConfigName": PropertyType(False, one_of(IS_STR, IS_DICT)),
        "DeploymentStyle": PropertyType(False, IS_DICT),
        "ServiceRoleArn": PropertyType(True, one_of(IS_STR, IS_DICT)),
        "TriggerConfigurations": PropertyType(False, is_type(list)),
    }

    runtime_attrs = {"name": lambda self: ref(self.logical_id)}



from samtranslator.model import PropertyType, Resource
from samtranslator.model.types import IS_DICT, IS_STR, list_of
from samtranslator.model.intrinsics import ref


class SNSSubscription(Resource):
    resource_type = "AWS::SNS::Subscription"
    property_types = {
        "Endpoint": PropertyType(True, IS_STR),
        "Protocol": PropertyType(True, IS_STR),
        "TopicArn": PropertyType(True, IS_STR),
        "Region": PropertyType(False, IS_STR),
        "FilterPolicy": PropertyType(False, IS_DICT),
        "RedrivePolicy": PropertyType(False, IS_DICT),
    }


class SNSTopicPolicy(Resource):
    resource_type = "AWS::SNS::TopicPolicy"
    property_types = {"PolicyDocument": PropertyType(True, IS_DICT), "Topics": PropertyType(True, list_of(str))}


class SNSTopic(Resource):
    resource_type = "AWS::SNS::Topic"
    property_types = {"TopicName": PropertyType(False, IS_STR)}
    runtime_attrs = {"arn": lambda self: ref(self.logical_id)}



from typing import Any, List, Optional

from samtranslator.model import PropertyType, Resource
from samtranslator.model.types import is_type, IS_STR
from samtranslator.utils.types import Intrinsicable


class Route53RecordSetGroup(Resource):
    resource_type = "AWS::Route53::RecordSetGroup"
    property_types = {
        "HostedZoneId": PropertyType(False, IS_STR),
        "HostedZoneName": PropertyType(False, IS_STR),
        "RecordSets": PropertyType(False, is_type(list)),
    }

    HostedZoneId: Optional[Intrinsicable[str]]
    HostedZoneName: Optional[Intrinsicable[str]]
    RecordSets: Optional[List[Any]]



import json
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union, TypeVar
from functools import partial

import pydantic
from pydantic import Extra, Field

# Value passed directly to CloudFormation; not used by SAM
PassThrough = Any  # TODO: Make it behave like typescript's unknown

# Intrinsic resolvable by the SAM transform
T = TypeVar("T")
SamIntrinsicable = Union[Dict[str, Any], T]
SamIntrinsic = Dict[str, Any]

# TODO: Get rid of this in favor of proper types
Unknown = Optional[Any]

DictStrAny = Dict[str, Any]

LenientBaseModel = pydantic.BaseModel

_thisdir = os.path.dirname(os.path.abspath(__file__))
_DOCS = json.loads(Path(_thisdir, "docs.json").read_bytes())


def get_prop(stem: str) -> Any:
    return partial(_get_prop, stem)


def _get_prop(stem: str, name: str) -> Any:
    docs = _DOCS["properties"][stem][name]
    return Field(
        title=name,
        description=docs,
        # https://code.visualstudio.com/docs/languages/json#_use-rich-formatting-in-hovers
        markdownDescription=docs,
    )


# By default strict: https://pydantic-docs.helpmanual.io/usage/model_config/#change-behaviour-globally
class BaseModel(LenientBaseModel):
    class Config:
        extra = Extra.forbid


# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html
class Ref(BaseModel):
    Ref: str



from samtranslator.model import PropertyType, Resource
from samtranslator.model.types import IS_DICT, is_type, IS_STR, list_of, one_of
from samtranslator.model.intrinsics import ref


class NestedStack(Resource):
    resource_type = "AWS::CloudFormation::Stack"
    # TODO: support passthrough parameters for stacks (Conditions, etc)
    property_types = {
        "TemplateURL": PropertyType(True, IS_STR),
        "Parameters": PropertyType(False, IS_DICT),
        "NotificationARNs": PropertyType(False, list_of(one_of(IS_STR, IS_DICT))),
        "Tags": PropertyType(False, list_of(IS_DICT)),
        "TimeoutInMinutes": PropertyType(False, is_type(int)),
    }

    runtime_attrs = {"stack_id": lambda self: ref(self.logical_id)}



from typing import Any, Dict, List, Optional

from samtranslator.model import PropertyType, Resource
from samtranslator.model.types import IS_DICT, list_of, IS_STR
from samtranslator.model.intrinsics import fnGetAtt, ref


class StepFunctionsStateMachine(Resource):
    resource_type = "AWS::StepFunctions::StateMachine"
    property_types = {
        "Definition": PropertyType(False, IS_DICT),
        "DefinitionString": PropertyType(False, IS_STR),
        "DefinitionS3Location": PropertyType(False, IS_DICT),
        "LoggingConfiguration": PropertyType(False, IS_DICT),
        "RoleArn": PropertyType(True, IS_STR),
        "StateMachineName": PropertyType(False, IS_STR),
        "StateMachineType": PropertyType(False, IS_STR),
        "Tags": PropertyType(False, list_of(IS_DICT)),
        "DefinitionSubstitutions": PropertyType(False, IS_DICT),
        "TracingConfiguration": PropertyType(False, IS_DICT),
    }

    Definition: Optional[Dict[str, Any]]
    DefinitionString: Optional[str]
    DefinitionS3Location: Optional[Dict[str, Any]]
    LoggingConfiguration: Optional[Dict[str, Any]]
    RoleArn: str
    StateMachineName: Optional[str]
    StateMachineType: Optional[str]
    Tags: Optional[List[Dict[str, Any]]]
    DefinitionSubstitutions: Optional[Dict[str, Any]]
    TracingConfiguration: Optional[Dict[str, Any]]

    runtime_attrs = {
        "arn": lambda self: ref(self.logical_id),
        "name": lambda self: fnGetAtt(self.logical_id, "Name"),
    }



from samtranslator.model import PropertyType, Resource
from samtranslator.model.types import IS_DICT, is_type, dict_of, list_of, IS_STR, one_of
from samtranslator.model.intrinsics import fnGetAtt, ref


class DynamoDBTable(Resource):
    resource_type = "AWS::DynamoDB::Table"
    property_types = {
        "AttributeDefinitions": PropertyType(True, list_of(IS_DICT)),
        "GlobalSecondaryIndexes": PropertyType(False, list_of(IS_DICT)),
        "KeySchema": PropertyType(False, list_of(IS_DICT)),
        "LocalSecondaryIndexes": PropertyType(False, list_of(IS_DICT)),
        "ProvisionedThroughput": PropertyType(False, dict_of(IS_STR, one_of(is_type(int), IS_DICT))),
        "StreamSpecification": PropertyType(False, IS_DICT),
        "TableName": PropertyType(False, one_of(IS_STR, IS_DICT)),
        "Tags": PropertyType(False, list_of(IS_DICT)),
        "SSESpecification": PropertyType(False, IS_DICT),
        "BillingMode": PropertyType(False, IS_STR),
    }

    runtime_attrs = {
        "name": lambda self: ref(self.logical_id),
        "arn": lambda self: fnGetAtt(self.logical_id, "Arn"),
        "stream_arn": lambda self: fnGetAtt(self.logical_id, "StreamArn"),
    }
